@@ -12,6 +12,7 @@ very fast. freeze_stable and flatten are usable for testing and analysis.
 Authors: Jean-Louis Fuchs <ganwell@fangorn.ch> https://github.com/ganwell
 
 Run doctests by with "python lib/freeze.py"
+
 """
 
 # Copyright (c) 2012, Adfinis SyGroup AG
@@ -41,11 +42,17 @@ Run doctests by with "python lib/freeze.py"
 
 
 import six
-import pickle
 import pprint
 import sys
+import json
+import gzip
 import difflib
 
+# These imports are need for doc tests.
+nothing = json.__file__
+nothing = gzip.__file__
+del nothing
+
 __all__ = [
     "freeze",
     "freeze_fast",
@@ -56,26 +63,21 @@ __all__ = [
     "tree_diff",
     "flatten",
     "frozen_equal_assert",
-    "pickle_protocol",
     "vformat",
-    "transparent_repr"
+    "transparent_repr",
+    "traverse_frozen_data",
+    "TraversalBasedReprCompare",
 ]
 
 
-# Setting the pickle protocol python2/3 compat
-try:
-    pickle_protocol = pickle.DEFAULT_PROTOCOL
-except:
-    pickle_protocol = pickle.HIGHEST_PROTOCOL
-
 try:
     _ignore_types = (six.string_types, tuple, bytes)
-except:
+except:  # pragma: no cover
     _ignore_types = (six.string_types, tuple)
 
 try:
     _string_types = (six.string_types,  bytes)
-except:
+except:  # pragma: no cover
     _string_types = (six.string_types)
 
 _primitive_types = (int, float, bool)
@@ -92,26 +94,29 @@ def freeze(data_structure, stringify=False):
     and will freeze any kind of python object.
 
     :param   data_structure: The structure to convert
-    :param        stringify: Stringify all leaves
+    :param        stringify: Stringify all non-primitive leaves
     :type         stringify: bool
 
-    >>> freeze([
+    We need freeze stable to test this across python versions. It doesn't make
+    too much sense, but at least there are some tests.
+
+    >>> freeze(freeze_stable([
     ...     'a',
     ...     [3, 4],
     ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
     ...     []
-    ... ])
-    ('a', (3, 4), (('a', (3, (('w', ((2, 3, 5), '3', 4)),))),), ())
+    ... ]))
+    ((), (((((('3', (2, 3, 5), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
 
-    >>> freeze([
+    >>> freeze(freeze_stable([
     ...     [],
     ...     'a',
     ...     [3, 4],
-    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
-    ... ])
-    ((), 'a', (3, 4), (('a', (3, (('w', ((2, 3, 5), '3', 4)),))),))
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,1,2])])}]},
+    ... ]))
+    ((), (((((('3', (1, 2, 3), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
 
-    >>> b"sdf" == freeze({3: b"sdf", 4: 245234534})[0][1]
+    >>> b"sdf" == freeze([[3, b"sdf"], [4, 245234534]])[0][1]
     True
 
     >>> testdata = json.loads(
@@ -121,13 +126,26 @@ def freeze(data_structure, stringify=False):
     >>> freeze(testdata) == freeze_fast(testdata)
     True
 
-    Hash anything!
+    # Testing if hasattr(data_structure, "__dict__")
 
-    >>> a = freeze(sys)
-    >>> len(a) > 0
+    >>> class A():
+    ...     pass
+    >>> a = A()
+    >>> b = freeze([a, {"other": a}])
+    >>> ("<class 'freeze.A'> at " in str(b) or
+    ... "<type 'instance'> at " in str(b))
     True
-    >>> hash(a) != None
+
+    >>> a = lambda a: a*a
+    >>> b = freeze([a, {"other": a}])
+    >>> ("<class 'function'> at " in str(b) or
+    ... "<type 'function'> at " in str(b))
     True
+
+    # Testing builtings
+
+    >>> freeze([dir])
+    ('<built-in function dir>',)
     """
 
     identity_set = set()
@@ -148,7 +166,12 @@ def freeze(data_structure, stringify=False):
             if tlen != -1:
                 # Except string and tuples
                 if not isinstance(data_structure, _ignore_types):
-                    return "%s at 0x%d" % (type(data_structure), idd)
+                    # I can't test this on python3. I would need a type that is
+                    # not handled above. Namespaces are the only thing I know
+                    # and python3 doesn't freeze namespaces :(
+                    return "%s at 0x%d" % (
+                        type(data_structure), idd
+                    )  # pragma: no cover
         else:
             identity_set.add(idd)
         # Dictize if possible (support objects)
@@ -164,7 +187,7 @@ def freeze(data_structure, stringify=False):
         # We don't freeze strings
         if not isinstance(data_structure, _string_types):
             tlen = -1
-            # If item has a length we frezze it
+            # If item has a length we freeze it
             try:
                 tlen = len(data_structure)
             except:
@@ -194,23 +217,33 @@ def freeze_fast(data_structure):
 
     :param   data_structure: The structure to convert
 
-    >>> freeze_fast([
+    We need freeze stable to test this across python versions. It doesn't make
+    too much sense, but at least there are some tests.
+
+    >>> freeze_fast(freeze_stable([
     ...     'a',
     ...     [3, 4],
     ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
     ...     []
-    ... ])
-    ('a', (3, 4), (('a', (3, (('w', ((2, 3, 5), '3', 4)),))),), ())
+    ... ]))
+    ((), (((((('3', (2, 3, 5), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
+
+    >>> freeze_fast(freeze_stable([
+    ...     [],
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ... ]))
+    ((), (((((('3', (2, 3, 5), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
 
     >>> freeze_fast([
     ...     [],
     ...     'a',
     ...     [3, 4],
-    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
-    ... ])
-    ((), 'a', (3, 4), (('a', (3, (('w', ((2, 3, 5), '3', 4)),))),))
+    ...     ['a', [3, ['w', [4, '3']]]]])
+    ((), 'a', (3, 4), ('a', (3, ('w', (4, '3')))))
 
-    >>> b"sdf" == freeze_fast({3: b"sdf", 4: 245234534})[0][1]
+    >>> b"sdf" == freeze_fast([[3, b"sdf"], [4, 245234534]])[0][1]
     True
     """
     def freeze_fast_helper(data_structure):
@@ -227,7 +260,7 @@ def freeze_fast(data_structure):
         # We don't freeze strings
         if not isinstance(data_structure, _string_types):
             tlen = -1
-            # If item has a length we frezze it
+            # If item has a length we freeze it
             try:
                 tlen = len(data_structure)
             except:
@@ -239,10 +272,11 @@ def freeze_fast(data_structure):
 
 
 def _recursive_sort(data_structure, assume_key=False):
+    """Will only work with stringified data_structures"""
     # We don't sort strings
     if not isinstance(data_structure, _string_types):
         tlen = -1
-        # If item has a length we frezze it
+        # If item has a length we sort it
         try:
             tlen = len(data_structure)
         except:
@@ -261,17 +295,19 @@ def _recursive_sort(data_structure, assume_key=False):
                         x,
                         assume_key=assume_key
                     ) for x in data_structure],
-                    key=lambda x: pickle.dumps(
-                        x,
-                        protocol=pickle_protocol
-                    )
+                    key=TraversalBasedReprCompare,
                 ))
     return data_structure
 
 
-def freeze_stable(data_structure, assume_key=False, stringify=False):
+def freeze_stable(data_structure, assume_key=False, stringify=True):
     """Like freeze but results are stable regarding sorting.
 
+    UNLIKE the other freeze variants stringify is set True by default.  Because
+    freeze_stable should only be used as a helper in testing, freeze_stable
+    uses the better and faster stringify method. If you actually have an
+    application other than testing feel free to set stringify to False.
+
     The stable functionality changes the datestructure (sorting)!
     For deterministic results don't use stable. If a two structures
     differ only in order / sorting of items freeze_stable will have
@@ -282,13 +318,11 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
     creates smaller structures for flatten and a smaller
     difference when comparing two flattened structures.
 
-    stringify allow to freeze_stable anything.
-
     :param   data_structure: The structure to convert
     :param       assume_key: Assume that substructures of len() == 2
                              are key-value pairs -> don't sort
     :type        assume_key: bool
-    :param        stringify: Stringify all leaves
+    :param        stringify: Stringify all non-primitive leaves
     :type         stringify: bool
 
 
@@ -297,27 +331,24 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
     ...     [3, 4],
     ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
     ...     []
-    ... ])
+    ... ], stringify=False)
     >>> a
-    ((), (((((((2, 3, 5), 4, '3'), 'w'),), 3), 'a'),), (3, 4), 'a')
+    ((), (((((('3', (2, 3, 5), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
 
     >>> b = freeze_stable([
     ...     [],
     ...     'a',
     ...     [4, 3],
     ...     {'a': [{'w' : set([4, '3', frozenset([3,5,2])])}, 3]},
-    ... ])
+    ... ], stringify=False)
     >>> b
-    ((), (((((((2, 3, 5), 4, '3'), 'w'),), 3), 'a'),), (3, 4), 'a')
+    ((), (((((('3', (2, 3, 5), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
 
     a is the same as b!!
 
     >>> a == b
     True
 
-    >>> b"sdf" == freeze_stable({3: b"sdf", 4: 245234534})[1][1]
-    True
-
     >>> a = freeze_stable([
     ...     'a',
     ...     [3, 4],
@@ -325,7 +356,7 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
     ...     []
     ... ], stringify=True)
     >>> a
-    ((), (((((((2, 3, 5), 4, '3'), 'w'),), 3), 'a'),), (3, 4), 'a')
+    ((), (((((('3', (2, 3, 5), 4), 'w'),), 3), 'a'),), 'a', (3, 4))
     >>> a = freeze_stable([
     ...     'a',
     ...     [5, 4],
@@ -333,7 +364,32 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
     ...     []
     ... ], stringify=True, assume_key=True)
     >>> a
-    ((), (5, 4), 'a', (('a', (3, (('w', ((2, 3, 5), 4, '3')),))),))
+    ((), 'a', (('a', (3, (('w', ('3', (2, 3, 5), 4)),))),), (5, 4))
+
+    # Testing if hasattr(data_structure, "__dict__")
+
+    >>> class A():
+    ...     pass
+    >>> a = A()
+    >>> b = freeze_stable([a, {"other": a}], stringify=False)
+    >>> ("<class 'freeze.A'> at " in str(b) or
+    ... "<type 'instance'> at " in str(b))
+    True
+
+    >>> a = lambda a: a*a
+    >>> b = freeze_stable([a, {"other": a}], stringify=False)
+    >>> ("<class 'function'> at " in str(b) or
+    ... "<type 'function'> at " in str(b))
+    True
+
+    # Testing builtings
+
+    >>> freeze_stable([dir], stringify=False)
+    ('<built-in function dir>',)
+
+
+    >>> freeze_stable({"a": "b"}, assume_key=True, stringify=False)
+    (('a', 'b'),)
     """
 
     if not stringify:
@@ -355,7 +411,12 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
             if tlen != -1:
                 # Except string and tuples
                 if not isinstance(data_structure, _ignore_types):
-                    return "%s at 0x%d" % (type(data_structure), idd)
+                    # I can't test this on python3. I would need a type that is
+                    # not handled above. Namespaces are the only thing I know
+                    # and python3 doesn't freeze namespaces :(
+                    return "%s at 0x%d" % (
+                        type(data_structure), idd
+                    )  # pragma: no cover
         else:
             identity_set.add(idd)
         # Dictize if possible (support objects)
@@ -371,7 +432,7 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
         # We don't freeze strings
         if not isinstance(data_structure, _string_types):
             tlen = -1
-            # If item has a length we frezze it
+            # If item has a length we freeze it
             try:
                 tlen = len(data_structure)
             except:
@@ -384,10 +445,7 @@ def freeze_stable(data_structure, assume_key=False, stringify=False):
                 else:
                     return tuple(sorted(
                         [freeze_stable_helper(x) for x in data_structure],
-                        key=lambda x: pickle.dumps(
-                            x,
-                            protocol=pickle_protocol
-                        )
+                        key=TraversalBasedReprCompare
                     ))
         # To guarantee that the result is hashable we do not return
         # builtin_function_or_method
@@ -414,7 +472,15 @@ def stable_hash(data_structure):
     Note: hash(freeze.freeze_fast(x)) is faster than freeze.recursive_hash(x)
     BUT freeze.stable_hash(x) is much faster than hash(freeze.stable_freeze(x)
 
-    :param   data_structure: The structure to hash"""
+    :param   data_structure: The structure to hash
+
+    >>> class A(object):
+    ...     def __init__(self):
+    ...         self.huhu = 53
+    >>> a = A()
+    >>> stable_hash([a, 4]) == stable_hash([4, a])
+    True
+    """
     # Dictize if possible (support objects)
     try:
         data_structure = data_structure.__dict__
@@ -428,7 +494,7 @@ def stable_hash(data_structure):
     # We don't iterate strings
     if not isinstance(data_structure, _string_types):
         tlen = -1
-        # If item has a length we frezze it
+        # If item has a length we freeze it
         try:
             tlen = len(data_structure)
         except:
@@ -442,7 +508,20 @@ def stable_hash(data_structure):
 def recursive_hash(data_structure):
     """Recursive hash of data-structure and objects. No cycle detection.
 
-    :param   data_structure: The structure to hash"""
+    :param   data_structure: The structure to hash
+
+    The hash is seeded different everytime python is started. So there is no
+    meaningful test yet.
+
+    >>> class A(object):
+    ...     def __init__(self):
+    ...         self.huhu = 53
+    >>> a = A()
+    >>> b = [a, 4]
+    >>> bool(recursive_hash(b))
+    True
+
+    """
     # Dictize if possible (support objects)
     try:
         data_structure = data_structure.__dict__
@@ -456,7 +535,7 @@ def recursive_hash(data_structure):
     # We don't iterate strings
     if not isinstance(data_structure, _string_types):
         tlen = -1
-        # If item has a length we frezze it
+        # If item has a length we freeze it
         try:
             tlen = len(data_structure)
         except:
@@ -467,6 +546,213 @@ def recursive_hash(data_structure):
     return hash(data_structure)
 
 
+def tree_diff(a, b, n=5, deterministic=True):
+    """Freeze and stringify any data-structure or object, traverse
+    it depth-first in-order and apply a unified diff.
+
+    Depth-first in-order is just like structure would be printed.
+
+    Annotation:
+
+    "(           x"       Going down to level: x
+    ")           x"       Going one level up from: x
+
+    :param             a: data_structure a
+    :param             b: data_structure b
+    :param             n: lines of context
+    :type              n: int
+
+    We need freeze stable to test this across python versions. It doesn't make
+    too much sense, but at least there are some tests.
+
+    >>> a = freeze_stable([
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ...     []
+    ... ])
+    >>> b = freeze_stable([
+    ...     'a',
+    ...     [7, 3],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
+    ...     []
+    ... ])
+    >>> transparent_repr("\\n".join(tree_diff(a, b).split("\\n")[2:]))
+    @@ -7,6 +7,6 @@
+           'w'),),
+         3),
+        'a'),),
+      'a',
+      (3,
+    -  4))
+    +  7))
+
+    >>> a = freeze_stable([
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, 'tree', frozenset([3,5,2])])}]},
+    ...     []
+    ... ])
+    >>> b = freeze_stable([
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
+    ...     []
+    ... ])
+    >>> transparent_repr("\\n".join(tree_diff(a, b).split("\\n")[2:]))
+    @@ -1,12 +1,12 @@
+     ((),
+    - 'a',
+    - (('a',
+    -   (((('tree',
+    + (((((('3',
+            (2,
+             3,
+             5),
+            4),
+           'w'),),
+    -    3)),),
+    +    3),
+    +   'a'),),
+    + 'a',
+      (3,
+       4))
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, 'tree', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> transparent_repr("\\n".join(
+    ...     tree_diff(a, b, deterministic=False).split("\\n")[2:]
+    ... ))
+    @@ -1,12 +1,12 @@
+     ((),
+    - 'a',
+    - (('a',
+    -   (((('tree',
+    + (((((('3',
+            (2,
+             3,
+             5),
+            4),
+           'w'),),
+    -    3)),),
+    +    3),
+    +   'a'),),
+    + 'a',
+      (3,
+       4))
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set(['3', 4, frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> transparent_repr("\\n".join(
+    ...     tree_diff(a, b, deterministic=False).split("\\n")[2:]
+    ... ))
+    <BLANKLINE>
+    """
+    a = freeze(a, stringify=True)
+    b = freeze(b, stringify=True)
+    if deterministic:
+        a = vformat(a).split("\n")
+        b = vformat(b).split("\n")
+    else:
+        a = vformat(_recursive_sort(a)).split("\n")
+        b = vformat(_recursive_sort(b)).split("\n")
+    return "\n".join(difflib.unified_diff(a, b, n=n, lineterm=""))
+
+
+def tree_diff_assert(a, b, n=5, deterministic=True):
+    """Assert if a equals b. Freeze and stringify any data-structure or object,
+    traverse it depth-first and apply a unified diff, to display the result.
+
+    :param             a: data_structure a
+    :param             b: data_structure b
+    :param             n: lines of context
+    :type              n: int
+    :param deterministic: Do not sort the tree
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, 'tree', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> try:
+    ...     tree_diff_assert(a, b, deterministic=False)
+    ... except:
+    ...     "GOT IT"
+    'GOT IT'
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set(['3', 4, frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> tree_diff_assert(a, b, deterministic=False)
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set(['3', 4, frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> try:
+    ...     tree_diff_assert(a, b, deterministic=True)
+    ... except:
+    ...     "GOT IT"
+    'GOT IT'
+
+    """
+
+    a = freeze(a, stringify=True)
+    b = freeze(b, stringify=True)
+    if deterministic:
+        a = vformat(a).split("\n")
+        b = vformat(b).split("\n")
+    else:
+        a = vformat(_recursive_sort(a)).split("\n")
+        b = vformat(_recursive_sort(b)).split("\n")
+    msg = "\n".join(difflib.unified_diff(a, b, n=n, lineterm=""))
+    if len(msg) != 0:
+        assert False, "difference: \n%s" % msg
+
+
 def _flatten_helper(iterable, pathlist, path, parent_index=False):
     # Do not iterate over strings
     if isinstance(iterable, _string_types):
@@ -492,30 +778,19 @@ def _flatten_helper(iterable, pathlist, path, parent_index=False):
         # If we have a tuple (length 2) we try to use
         # the first item as key
         if tlen == 2:
-            sublen = -1
-            try:
-                sublen = len(iterable[0])
-            except:
-                pass
-            if sublen == -1 or isinstance(iterable[0], _string_types):
-                # If the parent is an index, we don't need it since we
-                # have a key.
-                if parent_index:
-                    del path[-1]
-                _flatten_helper(
-                    iterable[1],
-                    pathlist,
-                    path + [str(iterable[0])]
-                )
-                return
+            # If the parent is an index, we don't need it since we
+            # have a key.
+            if parent_index:
+                del path[-1]
+            _flatten_helper(
+                iterable[1],
+                pathlist,
+                path + [str(iterable[0])]
+            )
+            return
         # We have a list
         index = 0
         for item in iterable:
-            sublen = -1
-            try:
-                sublen = len(item)
-            except:
-                pass
             _flatten_helper(
                 item,
                 pathlist,
@@ -525,113 +800,6 @@ def _flatten_helper(iterable, pathlist, path, parent_index=False):
             index += 1
 
 
-def tree_diff(a, b, n=5, deterministic=True):
-    """Freeze and stringify any data-structure or object, traverse
-    it depth-first in-order and apply a unified diff.
-
-    Depth-first in-order is just like structure would be printed.
-
-    Annotation:
-
-    "(           x"       Going down to level: x
-    ")           x"       Going one level up from: x
-
-    :param             a: data_structure a
-    :param             b: data_structure b
-    :param             n: lines of context
-    :type              n: int
-
-
-    >>> a = [
-    ...     'a',
-    ...     [3, 4],
-    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
-    ...     []
-    ... ]
-    >>> b = [
-    ...     'a',
-    ...     [4, 3],
-    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
-    ...     []
-    ... ]
-    >>> transparent_repr("\\n".join(tree_diff(a, b).split("\\n")[2:]))
-    @@ -1,8 +1,8 @@
-     ('a',
-    - (3,
-    -  4),
-    + (4,
-    +  3),
-      (('a',
-        (3,
-         (('w',
-           ((2,
-             3,
-
-    >>> a = [
-    ...     'a',
-    ...     [3, 4],
-    ...     {'a': [3, {'w' : set([4, 'tree', frozenset([3,5,2])])}]},
-    ...     []
-    ... ]
-    >>> b = [
-    ...     'a',
-    ...     [4, 3],
-    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
-    ...     []
-    ... ]
-    >>> transparent_repr("\\n".join(tree_diff(a, b).split("\\n")[2:]))
-    @@ -1,12 +1,12 @@
-     ('a',
-    - (3,
-    -  4),
-    + (4,
-    +  3),
-      (('a',
-        (3,
-         (('w',
-           ((2,
-             3,
-             5),
-    -       'tree',
-    +       '3',
-            4)),))),),
-      ())
-
-    """
-    a = freeze(a, stringify=True)
-    b = freeze(b, stringify=True)
-    if deterministic:
-        a = vformat(a).split("\n")
-        b = vformat(b).split("\n")
-    else:
-        a = vformat(_recursive_sort(a)).split("\n")
-        b = vformat(_recursive_sort(b)).split("\n")
-    return "\n".join(difflib.unified_diff(a, b, n=n, lineterm=""))
-
-
-def tree_diff_assert(a, b, n=5, deterministic=True):
-    """Assert if a equals b. Freeze and stringify any data-structure or object,
-    traverse it depth-first and apply a unified diff, to display the result.
-
-    :param             a: data_structure a
-    :param             b: data_structure b
-    :param             n: lines of context
-    :type              n: int
-    :param deterministic: Do not sort the tree
-    """
-    a = freeze(a, stringify=True)
-    b = freeze(b, stringify=True)
-    if deterministic:
-        a = vformat(a).split("\n")
-        b = vformat(b).split("\n")
-    else:
-        a = vformat(_recursive_sort(a)).split("\n")
-        b = vformat(_recursive_sort(b)).split("\n")
-    msg = "\n".join(difflib.unified_diff(a, b, n=n, lineterm=""))
-    if len(msg) != 0:
-        assert False, "difference: \n%s" % msg
-
-
 def flatten(data_structure, assume_key=True):
     """Converts a data-structure to a flat set of paths.
     It makes it easier to comapre data-structures in unittests.
@@ -655,13 +823,13 @@ def flatten(data_structure, assume_key=True):
     >>> flat_one = flatten(test_one)
     >>> vformat(flat_one)
     ['/0/()',
-     '/3/4',
-     '/2/a',
-     '/3/a/3/w/0/0/2',
-     '/3/a/3/w/0/1/3',
-     '/3/a/3/w/0/2/5',
-     '/3/a/3/w/1/4',
-     '/3/a/3/w/2/3']
+     '/1/a',
+     '/2/a/3/w/0/3',
+     '/2/a/3/w/1/0/2',
+     '/2/a/3/w/1/1/3',
+     '/2/a/3/w/1/2/5',
+     '/2/a/3/w/2/4',
+     '/3/4']
 
     >>> test_two = [
     ...    'a',
@@ -672,13 +840,13 @@ def flatten(data_structure, assume_key=True):
     >>> flat_two = flatten(test_two)
     >>> vformat(flat_two)
     ['/0/()',
-     '/3/4',
-     '/2/a',
-     '/3/a/3/w/0/0/2',
-     '/3/a/3/w/0/1/3',
-     '/3/a/3/w/0/2/5',
-     '/3/a/3/w/1/4',
-     '/3/a/3/w/2/3']
+     '/1/a',
+     '/2/a/3/w/0/3',
+     '/2/a/3/w/1/0/2',
+     '/2/a/3/w/1/1/3',
+     '/2/a/3/w/1/2/5',
+     '/2/a/3/w/2/4',
+     '/3/4']
 
      >>> len(set(flat_one) - set(flat_two))
      0
@@ -687,7 +855,8 @@ def flatten(data_structure, assume_key=True):
 
     _flatten_helper(freeze_stable(
         data_structure,
-        assume_key=assume_key
+        assume_key=assume_key,
+        stringify=True
     ), pathlist, [''])
     return pathlist
 
@@ -706,12 +875,81 @@ def frozen_equal_assert(a, b, deterministic=True):
     :param             a: data_structure a
     :param             b: data_structure b
     :param deterministic: use flatten only for creating the diff
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, 'tree', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> try:
+    ...     frozen_equal_assert(a, b, deterministic=False)
+    ... except:
+    ...     "GOT IT"
+    'GOT IT'
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4, 1],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3, 1],
+    ...     {'a': [3, {'w' : set(['3', 4, frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> frozen_equal_assert(a, b, deterministic=False)
+
+    >>> a = [
+    ...     'a',
+    ...     [3, 4],
+    ...     {'a': [3, {'w' : set([4, '3', frozenset([3,5,2])])}]},
+    ...     []
+    ... ]
+    >>> b = [
+    ...     'a',
+    ...     [4, 3],
+    ...     {'a': [3, {'w' : set(['3', 4, frozenset([2,5,3])])}]},
+    ...     []
+    ... ]
+    >>> try:
+    ...     frozen_equal_assert(a, b, deterministic=True)
+    ... except:
+    ...     "GOT IT"
+    'GOT IT'
+
+    Testing case where flatten doen't find a diff
+
+    >>> a = [
+    ...     '3',
+    ... ]
+    >>> b = [
+    ...     3,
+    ... ]
+    >>> try:
+    ...     frozen_equal_assert(a, b, deterministic=True)
+    ... except:
+    ...     "GOT IT"
+    'GOT IT'
+
     """
 
     if not deterministic:
-        freeze_func = lambda x: freeze_stable(x, assume_key=True)
+        freeze_func = lambda x: freeze_stable(
+            x,
+            assume_key=True,
+            stringify=True
+        )
     else:
-        freeze_func = freeze
+        freeze_func = lambda x: freeze(x, stringify=True)
     a = freeze_func(a)
     b = freeze_func(b)
     # If a == b we are save anyway
@@ -736,21 +974,13 @@ def frozen_equal_assert(a, b, deterministic=True):
                     )
 
 
-class TransparentRepr(object):
+class TransparentRepr(str):
     """That doctest calls __repr__ is very annoying, you can't follow PEP8
     on large objects. TransparentRepr makes string represent itself
     including all non-printable characters."""
-    def __init__(self, repr_str):
-        self.repr_str = repr_str
 
     def __repr__(self):
-        if isinstance(self.repr_str, six.string_types):
-            return self.repr_str
-        else:
-            return super(TransparentRepr, self).__repr__()
-
-    def split(self, *args, **kwargs):
-        return self.repr_str.split(*args, **kwargs)
+        return self
 
 
 def transparent_repr(string):
@@ -768,10 +998,96 @@ def vformat(*args, **kwargs):
     )
 
 
-if __name__ == "__main__":
+def traverse_frozen_data(data_structure):
+    """Yields the leaves of the frozen data-structure pre-order.
+
+    It will produce the same order as one would write data-structure."""
+    parent_stack = [data_structure]
+    while parent_stack:
+        node = parent_stack.pop(0)
+        # We don't iterate strings
+        tlen = -1
+        if not isinstance(node, _string_types):
+            # If item has a length we freeze it
+            try:
+                tlen = len(node)
+            except:
+                pass
+        if tlen == -1:
+            yield node
+        else:
+            parent_stack = list(node) + parent_stack
+
+
+class TraversalBasedReprCompare(object):
+    """Implements the comparison method for frozen data-structures based on
+    traverse_frozen_data.
+
+    >>> cm = TraversalBasedReprCompare
+    >>> cm(3) < cm(4)
+    True
+    >>> cm(4) > cm(3)
+    True
+    >>> cm(3) > cm(4)
+    False
+    >>> cm(3) == cm(3)
+    True
+    >>> cm(3) == cm(4)
+    False
+    >>> cm((3, 3)) > cm((3,))
+    True
+    >>> cm((3, 3)) == cm((3, 3))
+    True
+    >>> cm((3,)) > cm((3, 3))
+    False
+    >>> cm((3,)) == cm((3, 3))
+    False
+    """
+
+    def __init__(self, payload):
+        """Initialize a payload (usually a key) for comparison"""
+        self.payload   = payload
+
+    def _cmp(self, other):
+        """Generic cmp method to support python 2/3"""
+        #import pdb; pdb.set_trace()
+        self_gen  = traverse_frozen_data(self.payload)
+        other_gen = traverse_frozen_data(other.payload)
+        while True:
+            try:
+                self_node  = repr(next(self_gen))
+            except StopIteration:
+                self_node  = None
+            try:
+                other_node = repr(next(other_gen))
+            except StopIteration:
+                other_node = None
+            if self_node is None or other_node is None:
+                # We iterated to the end
+                if self_node is not None:
+                    return 1
+                if other_node is not None:
+                    return -1
+                return 0
+            if self_node != other_node:
+                return (
+                    self_node > other_node
+                ) - (
+                    self_node < other_node
+                )
+
+    def __lt__(self, other):
+        return self._cmp(other) < 0
+
+    def __eq__(self, other):
+        return self._cmp(other) == 0
+
+    def __cmp__(self, other):  # pragma: no cover
+        return self._cmp(other)
+
+
+if __name__ == "__main__":  # pragma: no cover
     import doctest
-    import json
-    import gzip
-    pickle_protocol = 2
     result = doctest.testmod()
     sys.exit(result.failed)
+# lint_ignore=W0702,R0912,R0911
